using MetagraphOptimization: insert_node!, insert_edge!, make_node, make_edge

_construction_string(::Incoming) = "Incoming()"
_construction_string(::Outgoing) = "Outgoing()"

_construction_string(::Electron) = "Electron()"
_construction_string(::Positron) = "Positron()"
_construction_string(::Photon) = "Photon()"

_construction_string(::PolX) = "PolX()"
_construction_string(::PolY) = "PolY()"
_construction_string(::SpinUp) = "SpinUp()"
_construction_string(::SpinDown) = "SpinDown()"

function _parse_particle(name::String)
    local dir
    if startswith(name, "inc_")
        dir = Incoming()
    elseif startswith(name, "out_")
        dir = Outgoing()
    else
        throw(InvalidInputError("failed to parse particle direction from \"$name\""))
    end

    name = name[4:end]

    local species
    if startswith(name, "el")
        species = Electron()
    elseif startswith(name, "ph")
        species = Photon()
    elseif startswith(name, "po")
        species = Positron()
    else
        throw(InvalidInputError("failed to parse particle species from name \"$name\""))
    end

    name = name[3:end]

    local spin_pol
    if startswith(name, "su")
        spin_pol = SpinUp()
    elseif startswith(name, "sd")
        spin_pol = SpinDown()
    elseif startswith(name, "px")
        spin_pol = PolX()
    elseif startswith(name, "py")
        spin_pol = PolY()
    else
        throw(
            InvalidInputError(
                "failed to parse particle spin or polarization from \"$name\""
            ),
        )
    end

    name = name[3:end]

    index = parse(Int, name)
    return (dir, species, spin_pol, index)
end

function input_expr(instance::GenericQEDProcess, name::String, psp_symbol::Symbol)
    if startswith(name, "bs_")
        (dir, species, spin_pol, index) = _parse_particle(name[4:end])
        dir_str = _construction_string(dir)
        species_str = _construction_string(species)
        sp_str = _construction_string(spin_pol)

        return Meta.parse(
            "BaseStateInput(
                ParticleStateful($dir_str, $species_str, momentum($psp_symbol, $dir_str, $species_str, $index)),
                $sp_str,
            )",
        )

    elseif startswith(name, "pr_")
        index = parse(Int, name[4:end]) # get index of the virtual particle in the process

        vp = virtual_particles(proc)[index]
        return Meta.parse("PropagatorInput(
                              VirtualParticle(
                                process($psp_symbol),
                                $species_str,
                                $(vp.in_particle_contributions),
                                $(vp.out_particle_contributions)
                              ),
                              $psp_symbol
                          )")
    else
        throw(InvalidInputError("failed to parse node name \"$name\""))
    end
end

_species_str(::Photon) = "ph"
_species_str(::Electron) = "el"
_species_str(::Positron) = "po"

_spin_pol_str(::SpinUp) = "su"
_spin_pol_str(::SpinDown) = "sd"
_spin_pol_str(::PolX) = "px"
_spin_pol_str(::PolY) = "py"

_dir_str(::Incoming) = "inc"
_dir_str(::Outgoing) = "out"

_spin_pols(::AllSpin) = (SpinUp(), SpinDown())
_spin_pols(::SpinUp) = (SpinUp(),)
_spin_pols(::SpinDown) = (SpinDown(),)
_spin_pols(::AllPol) = (PolX(), PolY())
_spin_pols(::PolX) = (PolX(),)
_spin_pols(::PolY) = (PolY(),)

_is_external(p::VirtualParticle) = number_contributions(p) == 1

function _base_state_name(p::VirtualParticle)
    proc = process(p)

    dir = sum(in_contributions(p)) == 1 ? Incoming() : Outgoing()

    # find particle in the contributions
    index = 0
    contribs = is_incoming(dir) ? in_contributions(p) : out_contributions(p)
    for p in particles(proc, dir)
        index += 1
        if contribs[index]
            break
        end
    end

    species = particles(proc, dir)[index]

    # find particle index of *this species*
    species_index = 0
    for i in 1:index
        if particles(proc, dir)[i] == species
            species_index += 1
        end
    end

    spin_pol = spin_or_pol(proc, dir, species, species_index)

    return string.(
        "bs_$(_dir_str(dir))_$(_species_str(species))_",
        _spin_pol_str.(_spin_pols(spin_pol)),
        "_$(species_index)",
    )
end

function generate_DAG(proc::GenericQEDProcess)
    external_particles = _pseudo_virtual_particles(proc) # external particles that will be input to base_state tasks
    particles = virtual_particles(proc)                  # virtual particles that will be input to propagator tasks
    pairs = particle_pairs(particles)                    # pairs to generate the pair tasks
    triples = total_particle_triples(particles)          # triples to generate the triple tasks

    graph = DAG()

    # -- Base State Tasks --
    base_state_task_outputs = Dict()
    for dir in (Incoming(), Outgoing())
        for species in (Electron(), Positron(), Photon())
            for index in 1:number_particles(proc, dir, species)
                for spin_pol in _spin_pols(spin_or_pol(proc, dir, species, index))
                    # gen entry nodes
                    # names are "bs_<dir>_<species>_<spin/pol>_<index>"
                    data_node_name = "bs_$(_dir_str(dir))_$(_species_str(species))_$(_spin_pol_str(spin_pol))_$(index)"

                    data_in = insert_node!(
                        graph,
                        make_node(DataTask(0), data_node_name);
                        track=false,
                        invalidate_cache=false,
                    )

                    # generate initial base_state tasks
                    compute_base_state = insert_node!(
                        graph,
                        make_node(ComputeTask_BaseState());
                        track=false,
                        invalidate_cache=false,
                    )

                    data_out = insert_node!(
                        graph, make_node(DataTask(0)); track=false, invalidate_cache=false
                    )

                    insert_edge!(graph, data_in, compute_base_state)
                    insert_edge!(graph, compute_base_state, data_out)

                    base_state_task_outputs[data_node_name] = data_out
                end
            end
        end
    end

    # -- Propagator Tasks --
    propagator_task_outputs = Dict()
    vp_index = 0
    for vp in virtual_particles(proc)
        vp_index += 1

        data_node_name = "pr_$vp_index"

        data_in = insert_node!(
            graph, make_node(DataTask(0)); track=false, invalidate_cache=false
        )
        compute_vp_propagator = insert_node!(
            graph, make_node(ComputeTask_Propagator()); track=false, invalidate_cache=false
        )
        data_out = insert_node!(
            graph, make_node(DataTask(0)); track=false, invalidate_cache=false
        )

        insert_edge!(graph, data_in, compute_vp_propagator)
        insert_edge!(graph, compute_vp_propagator, data_out)

        propagator_task_outputs[data_node_name] = data_out
    end

    # -- Pair Tasks --
    pair_task_outputs = Dict()
    for (product_particle, input_particle_vector) in pairs
        # for all spins/pols of particles in product_particles do ...

        pair_task_outputs[product_particle] = Vector()

        for input_particles in input_particle_vector
            particles_data_out_nodes = (Vector(), Vector())
            c = 0
            for p in input_particles
                c += 1
                if (is_external(p))
                    # grab from base_states (broadcast over _base_state_name because it is a tuple for different spin_pols)
                    push!.(
                        Ref(particles_data_out_nodes[c]),
                        getindex.(Ref(base_state_task_outputs), _base_state_name(p)),
                    )
                else
                    # grab from propagated particles
                    push!(particles_date_out_nodes[c], pair_task_outputs[p])
                end
            end

            for in_nodes in Iterators.product(input_particles...)
                # make the compute pair nodes for every combination of the found input_particle_nodes to get all spin/pol combinations

                #insert_node!(graph, )

            end
            # make the collect pair and propagate nodes

        end

        data_out_propagated = insert_node!(
            graph, make_node(DataTask(0)); track=false, invalidate_caches=false
        )

        pair_task_outputs[p] = data_out_propagated

        # ... end do
    end

    return graph
end
